import os

import click
from click_repl import repl

import hokusai
from hokusai.lib.command import command
from hokusai.lib.common import set_verbosity, CONTEXT_SETTINGS

@click.group()
def base(context_settings=CONTEXT_SETTINGS):
  """Hokusai is a CLI for managing application deployments on Kubernetes"""
  pass

@base.command(context_settings=CONTEXT_SETTINGS)
@click.pass_context
def tree(ctx):
  """Print a tree of available commands"""
  hokusai.print_command_tree(ctx.find_root().command)

@base.command(context_settings=CONTEXT_SETTINGS)
def console():
    """Start an interactive console session"""
    repl(click.get_current_context())


@base.command(context_settings=CONTEXT_SETTINGS)
@click.option('--kubeconfig-dir', type=click.STRING, default=None, help='Directory to install kubeconfig into. The kubeconfig file will be named "config". (default: None)')
@click.option('--kubectl-dir', type=click.STRING, default=None, help='Directory to install kubectl into. The executable will be named "kubectl". (default: None)')
@click.option('--skip-kubeconfig', type=click.BOOL, is_flag=True, help='Skip kubeconfig install.')
@click.option('--skip-kubectl', type=click.BOOL, is_flag=True, help='Skip kubectl install.')
@click.option('-v', '--verbose', type=click.BOOL, is_flag=True, help='Verbose output.')
def configure(kubeconfig_dir, kubectl_dir, skip_kubeconfig, skip_kubectl, verbose):
  """Pull new Hokusai global config, download kubeconfig, install kubectl, save final global config to ~/.hokusai.yml"""
  set_verbosity(verbose)
  command(
    hokusai.hokusai_configure,
    kubeconfig_dir,
    kubectl_dir,
    skip_kubeconfig,
    skip_kubectl,
    config_check=False
  )


@base.command(context_settings=CONTEXT_SETTINGS)
@click.option('--project-name', type=click.STRING, default=os.path.basename(hokusai.CWD), help='The project name (default: name of current directory)')
@click.option('--template-remote', type=click.STRING, help='Git remote of templates to use - you can specify a branch via <git-remote>#<branch>')
@click.option('--template-dir', type=click.STRING, help='Directory of templates to use - can be used with --template-remote')
@click.option('--var', type=click.STRING, multiple=True, help='Extra variables to render Jinja templates in the form of key=value')
@click.option('--allow-missing-vars', is_flag=True, help='Do not fail on undefined template vars')
@click.option('-v', '--verbose', type=click.BOOL, is_flag=True, help='Verbose output')
def setup(project_name, template_remote, template_dir, var, allow_missing_vars, verbose):
  """Set up Hokusai for the current project"""
  set_verbosity(verbose)
  command(
    hokusai.setup,
    project_name,
    template_remote,
    template_dir,
    var,
    allow_missing_vars,
    config_check=False
  )


@base.command(context_settings=CONTEXT_SETTINGS)
@click.option('-f', '--filename', type=click.STRING, help='Use the given docker-compose Yaml file (default: ./hokusai/build.yml)')
@click.option('-v', '--verbose', type=click.BOOL, is_flag=True, help='Verbose output')
def build(filename, verbose):
  """Build the Docker image defined in ./hokusai/build.yml"""
  set_verbosity(verbose)
  command(
    hokusai.build,
    filename
  )


@base.command(context_settings=CONTEXT_SETTINGS)
@click.option('--build/--no-build', default=True, help='Force a build of the :latest image before running the test suite (default: true)')
@click.option('--cleanup/--no-cleanup', default=False, help='Remove containers on exit / error (default: False)')
@click.option('-f', '--filename', type=click.STRING, help='Use the given docker-compose Yaml file (default: ./hokusai/test.yml)')
@click.option('--service-name', type=click.STRING, help="The service name to treat as the test container (default: the value of 'project-name' in `hokusai/config.yml`)")
@click.option('-v', '--verbose', type=click.BOOL, is_flag=True, help='Verbose output')
def test(build, cleanup, filename, service_name, verbose):
  """Boot the docker-compose test environment defined by `./hokusai/test.yml` and run the test suite

  Return the exit code of the container with the name 'project-name' in `hokusai/config.yml`"""
  set_verbosity(verbose)
  command(
    hokusai.test,
    build,
    cleanup,
    filename,
    service_name
  )


@base.command(context_settings=CONTEXT_SETTINGS)
@click.option('-v', '--verbose', type=click.BOOL, is_flag=True, help='Verbose output')
def check(verbose):
  """Check Hokusai dependencies and configuration"""
  set_verbosity(verbose)
  command(
    hokusai.check
  )


@base.command(context_settings=CONTEXT_SETTINGS)
def version():
  """Print Hokusai's version and exit"""
  command(
    hokusai.version,
    config_check=False
  )

if __name__ == '__main__':
  base(obj={})



import os

import json

from hokusai.lib.common import shout, write_temp_file
from hokusai.lib.config import HOKUSAI_TMP_DIR
from hokusai.lib.exceptions import HokusaiError
from hokusai.services.kubectl import Kubectl


class Namespace:
  ''' represent a Kubernetes namespace '''
  def __init__(self, context, name, labels={}):
    self.context = context
    self.kctl = Kubectl(context)
    self.name = name
    self.labels = labels
    self.struct = {}

  def create(self):
    ''' create namespace '''
    if self.name == 'default':
      raise HokusaiError(
        f'Cannot create "default" namespace.'
      )
    metadata = {
      'name': self.name,
      'labels': self.labels
    }
    self.struct = {
      'apiVersion': 'v1',
      'kind': 'Namespace',
      'metadata': metadata
    }
    path = write_temp_file(
      json.dumps(self.struct), HOKUSAI_TMP_DIR
    )
    self.kctl.apply(path)

  def delete(self):
    ''' delete namespace '''
    if self.name == 'default':
      raise HokusaiError(f'Cannot delete "default" namespace.')
    shout(
      self.kctl.command(f'delete namespace {self.name}'),
      print_output=True
    )



import boto3

from urllib.parse import urlparse

# blanket import to avoid circular import error
import hokusai.lib.common as common


class S3Interface:
  ''' interface with AWS S3 '''
  def __init__(self):
    self._client = boto3.client(
      's3', region_name=common.get_region_name()
    )

  def download(self, uri, target_file):
    ''' download s3://bucket/foo/bar to target_file '''
    parsed_uri = urlparse(uri)
    bucket_name = parsed_uri.netloc
    key_name = parsed_uri.path.lstrip('/')
    if common.get_verbosity():
      common.print_green(
        f'Downloading {uri} to {target_file} ...',
        newline_after=True
      )
    try:
      self._client.download_file(
        bucket_name, key_name, target_file
      )
    except:
      common.print_red(
        f'Error: Failed to download {uri} to {target_file}'
      )
      raise



import os
import sys
import traceback

from hokusai.lib.common import print_red
from hokusai.lib.exceptions import CalledProcessError, HokusaiError
from hokusai.lib.config import config

def command(container_command, *args, config_check=True, **kwargs):
  try:
    if config_check:
      config.check()
    result = container_command(*args, **kwargs)
    if result is None:
      sys.exit(0)
    else:
      sys.exit(result)
  except SystemExit:
    raise
  except KeyboardInterrupt:
    raise
  except HokusaiError as e:
    print_red("ERROR: %s" % e.message)
    sys.exit(e.return_code)
  except CalledProcessError as e:
    print_red(traceback.format_exc())
    if hasattr(e, 'output') and e.output is not None:
      if type(e.output) == bytes:
        print_red("ERROR: %s" % e.output.decode('utf-8'))
      else:
        print_red("ERROR: %s" % e.output)
    sys.exit(e.returncode)
  except Exception as e:
    print_red(traceback.format_exc())
    sys.exit(1)



import click

import hokusai

from hokusai.cli.base import base
from hokusai.lib.command import command
from hokusai.lib.common import set_verbosity, CONTEXT_SETTINGS, confirm_delete_resources
from hokusai.lib.config import config

KUBE_CONTEXT = 'staging'

@base.group()
def staging(context_settings=CONTEXT_SETTINGS):
  """Interact with the staging Kubernetes environment
  defined by the `staging` context in `~/.kube/config` and the 
  Kubernetes resources in `./hokusai/staging.yml`"""
  pass

@staging.command(context_settings=CONTEXT_SETTINGS)
@click.option('-f', '--filename', type=click.STRING, help='Use the given Kubernetes Yaml file (default ./hokusai/staging.yml)')
@click.option('-e', '--environment', type=click.STRING, multiple=True, help='Create stack with the given environment variables (only applies if --filename is not supplied)')
@click.option('-v', '--verbose', type=click.BOOL, is_flag=True, help='Verbose output')
def create(filename, environment, verbose):
  """Create the Kubernetes resources defined in ./hokusai/staging.yml"""
  set_verbosity(verbose)
  command(
    hokusai.k8s_create,
    KUBE_CONTEXT,
    filename=filename,
    environment=environment
  )


@staging.command(context_settings=CONTEXT_SETTINGS)
@click.option('-f', '--filename', type=click.STRING, help='Use the given Kubernetes Yaml file (default ./hokusai/staging.yml)')
@click.option('-v', '--verbose', type=click.BOOL, is_flag=True, help='Verbose output')
@click.option('-y', '--assume-yes', type=click.BOOL, is_flag=True, help='Assume yes to confirmation prompt')
def delete(filename, verbose, assume_yes):
  """Delete the Kubernetes resources defined in ./hokusai/staging.yml"""
  set_verbosity(verbose)
  
  # Skip confirmation if --assume-yes is used
  if not assume_yes:
    if not confirm_delete_resources(config.project_name, KUBE_CONTEXT, filename):
      return
  
  command(
    hokusai.k8s_delete,
    KUBE_CONTEXT,
    filename=filename
  )


@staging.command(context_settings=CONTEXT_SETTINGS)
@click.option('--check-branch', type=click.STRING, default="main", help='Check branch before updating (default: main)')
@click.option('--check-remote', type=click.STRING, help='Check remotes before updating (otherwise check all remotes)')
@click.option('--skip-checks', type=click.BOOL, is_flag=True, help='Skip all checks and update configuration recklessly')
@click.option('-f', '--filename', type=click.STRING, help='Use the given Kubernetes Yaml file (default ./hokusai/staging.yml)')
@click.option('--dry-run', type=click.BOOL, is_flag=True, help='Perform a dry run of the configuration update')
@click.option('-v', '--verbose', type=click.BOOL, is_flag=True, help='Verbose output')
def update(check_branch, check_remote, skip_checks, filename, dry_run, verbose):
  """Update the Kubernetes resources defined in ./hokusai/staging.yml"""
  set_verbosity(verbose)
  command(
    hokusai.k8s_update,
    KUBE_CONTEXT,
    check_branch=check_branch,
    check_remote=check_remote,
    skip_checks=skip_checks,
    filename=filename,
    dry_run=dry_run
  )


@staging.command(context_settings=CONTEXT_SETTINGS)
@click.option('--resources/--no-resources', default=True, help='Print Kubernetes API objects defined in ./hokusai/staging.yml (default: true)')
@click.option('--pods/--no-pods', default=True, help='Print pods (default: true)')
@click.option('--describe', type=click.BOOL, is_flag=True, help="Print 'kubectl describe' output for resources and pods")
@click.option('--top', type=click.BOOL, is_flag=True, help='Print top pods')
@click.option('-f', '--filename', type=click.STRING, help='Use the given Kubernetes Yaml file (default ./hokusai/staging.yml)')
@click.option('-v', '--verbose', type=click.BOOL, is_flag=True, help='Verbose output')
def status(resources, pods, describe, top, filename, verbose):
  """Print Kubernetes resources in the staging context"""
  set_verbosity(verbose)
  command(
    hokusai.k8s_status,
    KUBE_CONTEXT,
    resources,
    pods,
    describe,
    top,
    filename=filename
  )


@staging.command(context_settings=CONTEXT_SETTINGS)
@click.argument('container_command', type=click.STRING)
@click.option('--tty', type=click.BOOL, is_flag=True, help='Attach the terminal')
@click.option('--tag', type=click.STRING, help='The image tag to run (defaults to "staging")')
@click.option('--env', type=click.STRING, multiple=True, help='Environment variables in the form of "KEY=VALUE"')
@click.option('--constraint', type=click.STRING, multiple=True, help='Constrain command to run on nodes matching labels in the form of "key=value"')
@click.option('-v', '--verbose', type=click.BOOL, is_flag=True, help='Verbose output')
def run(container_command, tty, tag, env, constraint, verbose):
  """Launch a new container and run a command"""
  set_verbosity(verbose)
  command(
    hokusai.run,
    KUBE_CONTEXT,
    container_command,
    tty,
    tag,
    env,
    constraint
  )


@staging.command(context_settings=CONTEXT_SETTINGS)
@click.option('-s', '--timestamps', type=click.BOOL, is_flag=True, help='Include timestamps')
@click.option('-f', '--follow', type=click.BOOL, is_flag=True, help='Follow logs')
@click.option('-t', '--tail', type=click.INT, help="Number of lines of recent logs to display")
@click.option('-p', '--previous', type=click.BOOL, is_flag=True, help='Get from the previous run of the Pod container(s)')
@click.option('-l', '--label', type=click.STRING, multiple=True, help='Filter pods by additional label selectors in the form of "key=value"')
@click.option('-v', '--verbose', type=click.BOOL, is_flag=True, help='Verbose output')
def logs(timestamps, follow, tail, previous, label, verbose):
  """Get container logs"""
  set_verbosity(verbose)
  command(
    hokusai.logs,
    KUBE_CONTEXT,
    timestamps,
    follow,
    tail,
    previous,
    label
  )

@staging.command(context_settings=CONTEXT_SETTINGS)
@click.argument('tag', type=click.STRING)
@click.option('--migration', type=click.STRING, help='Run a migration before deploying')
@click.option('--constraint', type=click.STRING, multiple=True, help='Constrain migration and deploy hooks to run on nodes matching labels in the form of "key=value"')
@click.option('--git-remote', type=click.STRING, help='Push deployment tags to git remote')
@click.option('-t', '--timeout', type=click.INT, default=600, help="Timeout deployment rollout after N seconds (default 600)")
@click.option('-u', '--update-config', type=click.BOOL, is_flag=True, help='Also update Kubernetes config')
@click.option('-f', '--filename', type=click.STRING, help='If updating config, use the Kubernetes Yaml file in the ./hokusai directory (default staging.yml)')
@click.option('-v', '--verbose', type=click.BOOL, is_flag=True, help='Verbose output')
def deploy(tag, migration, constraint, git_remote, timeout, update_config, filename, verbose):
  """Update the project's deployment(s) to reference
  the given image tag and update the tag staging
  to reference the same image"""
  set_verbosity(verbose)
  command(
    hokusai.update,
    KUBE_CONTEXT,
    tag,
    migration,
    constraint,
    git_remote,
    timeout,
    update_config=update_config,
    filename=filename
  )


@staging.command(context_settings=CONTEXT_SETTINGS)
@click.option('-d', '--deployment', type=click.STRING, help='Only refresh the given deployment')
@click.option('-v', '--verbose', type=click.BOOL, is_flag=True, help='Verbose output')
def refresh(deployment, verbose):
  """Refresh the project's deployment(s) by recreating the currently running containers"""
  set_verbosity(verbose)
  command(
    hokusai.refresh,
    KUBE_CONTEXT,
    deployment
  )


@staging.command(context_settings=CONTEXT_SETTINGS)
@click.option('-d', '--deployment', type=click.STRING, help='Only refresh the given deployment')
@click.option('-v', '--verbose', type=click.BOOL, is_flag=True, help='Verbose output')
def restart(deployment, verbose):
  """Alias for 'refresh'"""
  set_verbosity(verbose)
  command(
    hokusai.refresh,
    KUBE_CONTEXT,
    deployment
  )


@staging.group()
def env(context_settings=CONTEXT_SETTINGS):
  """Interact with the runtime environment for the application"""
  pass


@env.command(context_settings=CONTEXT_SETTINGS)
@click.argument('env_vars', type=click.STRING, nargs=-1)
@click.option('-v', '--verbose', type=click.BOOL, is_flag=True, help='Verbose output')
def get(env_vars, verbose):
  """Print environment variables stored on the Kubernetes server"""
  set_verbosity(verbose)
  command(
    hokusai.get_env,
    KUBE_CONTEXT,
    env_vars
  )


@env.command(context_settings=CONTEXT_SETTINGS)
@click.argument('env_vars', type=click.STRING, nargs=-1)
@click.option('-v', '--verbose', type=click.BOOL, is_flag=True, help='Verbose output')
def set(env_vars, verbose):
  """Set environment variables - each of {ENV_VARS} must be in of form 'KEY=VALUE'"""
  set_verbosity(verbose)
  command(
    hokusai.set_env,
    KUBE_CONTEXT,
    env_vars
  )


@env.command(context_settings=CONTEXT_SETTINGS)
@click.argument('env_vars', type=click.STRING, nargs=-1)
@click.option('-v', '--verbose', type=click.BOOL, is_flag=True, help='Verbose output')
def unset(env_vars, verbose):
  """Unset environment variables - each of {ENV_VARS} must be of the form 'KEY'"""
  set_verbosity(verbose)
  command(
    hokusai.unset_env,
    KUBE_CONTEXT,
    env_vars
  )



import os

import yaml

from hokusai import CWD
from hokusai.lib.common import get_verbosity, print_green, print_yellow, shout
from hokusai.lib.config import HOKUSAI_CONFIG_DIR
from hokusai.lib.exceptions import CalledProcessError
from hokusai.lib.template_selector import TemplateSelector
from hokusai.services.yaml_spec import YamlSpec


def detect_compose_command():
  ''' decide what command to use for Docker Compose '''
  command_to_use = ''
  try:
    shout('which docker-compose')
    if get_verbosity():
      print_green('Found docker-compose.')
    command_to_use = 'docker-compose'
  except CalledProcessError:
    if get_verbosity():
      print_yellow(
        'docker-compose command not found. ' +
        'Will use "docker compose" assuming it exists.'
      )
    command_to_use = 'docker compose'
  return command_to_use

def follow_extends(docker_compose_yml):
  with open(docker_compose_yml, 'r') as f:
    rendered_templates = []
    struct = yaml.safe_load(f.read())
    for service_name, service_spec in list(struct['services'].items()):
      if 'extends' not in service_spec or 'file' not in service_spec['extends']:
        continue
      extended_filename = service_spec['extends']['file']
      extended_template_path = os.path.join(CWD, HOKUSAI_CONFIG_DIR, extended_filename)
      if not os.path.isfile(extended_template_path):
        extended_template_path = os.path.join(CWD, HOKUSAI_CONFIG_DIR, extended_filename + '.j2')
      extended_template = TemplateSelector().get(extended_template_path)
      rendered_templates.append(YamlSpec(extended_template).to_file())
    return rendered_templates

def generate_compose_command(filename, default_yaml_file):
  ''' return Docker Compose command '''
  docker_compose_yml = render_docker_compose_yml(
    filename,
    default_yaml_file
  )
  # docker-compose v2 switched to using '-' as separator in image name,
  # resulting in 'hokusai-<project>'
  # COMPOSE_COMPATIBILITY=true forces v2 to use '_',
  # resulting in 'hokusai_<project>', matching v1
  return (
    'COMPOSE_COMPATIBILITY=true ' +
    detect_compose_command() +
    f' -f {docker_compose_yml}'
  )

def get_yaml_template(filename, default_yaml_file):
  ''' return yaml template '''
  if filename is None:
    yaml_template = TemplateSelector().get(
      os.path.join(
        CWD,
        HOKUSAI_CONFIG_DIR,
        default_yaml_file
      )
    )
  else:
    yaml_template = TemplateSelector().get(filename)
  return yaml_template

def render_docker_compose_yml(filename, default_yaml_file):
  ''' return path of rendered Docker Compose yaml '''
  yaml_template = get_yaml_template(filename, default_yaml_file)
  docker_compose_yml = YamlSpec(yaml_template).to_file()
  follow_extends(docker_compose_yml)
  return docker_compose_yml



import os

import datetime
import json
import time
import yaml

from hokusai import CWD
from hokusai.lib.common import (
  print_green,
  print_red,
  print_yellow,
  shout,
  shout_concurrent,
  write_temp_file,
  yaml_content_with_header
)
from hokusai.lib.config import HOKUSAI_CONFIG_DIR, HOKUSAI_TMP_DIR, config
from hokusai.lib.exceptions import CalledProcessError, HokusaiError
from hokusai.lib.template_selector import TemplateSelector
from hokusai.services.command_runner import CommandRunner
from hokusai.services.ecr import ECR, ClientError
from hokusai.services.kubectl import Kubectl
from hokusai.services.yaml_spec import YamlSpec


class Deployment:
  def __init__(self, context, deployment_name=None, namespace=None):
    self.context = context
    self.namespace = namespace
    self.kctl = Kubectl(self.context, namespace=namespace)
    self.ecr = ECR()
    if deployment_name:
      self.cache = [self.kctl.get_object("deployment %s" % deployment_name)]
    else:
      self.cache = self.kctl.get_objects('deployment', selector="app=%s,layer=application" % config.project_name)

  def update(
    self,
    tag,
    constraint,
    git_remote,
    timeout,
    update_config=False,
    filename=None,
    render_template=True
  ):
    if not self.ecr.project_repo_exists():
      raise HokusaiError("Project repo does not exist.  Aborting.")

    digest = self.ecr.image_digest_for_tag(tag)
    if digest is None:
      raise HokusaiError("Could not find an image digest for tag %s.  Aborting." % tag)
    digest_half = digest[len(digest)//2:]

    if self.namespace is None:
      print_green("Deploying %s to %s..." % (digest, self.context), newline_after=True)
    else:
      print_green("Deploying %s to %s/%s..." % (digest, self.context, self.namespace), newline_after=True)

    """
    This logic should be refactored, but essentially if namespace and filename are provided, the caller is
    a review app, while if namespace is None it is either staging or production.  If filename is unset for staging
    or production it is targeting the 'canonical' app, i.e. staging.yml or production.yml while if it is set it is
    trageting a 'canary' app.

    For the canonical app, run deploy hooks and post-depoy steps creating deployment tags
    For a canary app, skip deploy hooks and post-deploy steps
    For review apps, run deploy hooks but skip post-deploy steps

    For all deployment rollouts, if update_config or filename targets a yml file, bust the
    deployment cache using k8s field selectors and get deployments to watch the rollout from
    the yml file spec
    """

    # Run the pre-deploy hook for the canonical app or a review app
    if config.pre_deploy and (filename is None or (filename and self.namespace)):
      print_green("Running pre-deploy hook '%s'..." % config.pre_deploy, newline_after=True)
      return_code = CommandRunner(self.context, namespace=self.namespace).run(digest, config.pre_deploy, constraint=constraint, tty=False)
      if return_code:
        raise HokusaiError("Pre-deploy hook failed with return code %s" % return_code, return_code=return_code)

    if filename is None:
      yaml_template = TemplateSelector().get(os.path.join(CWD, HOKUSAI_CONFIG_DIR, self.context))
    else:
      yaml_template = TemplateSelector().get(filename)

    yaml_spec = YamlSpec(yaml_template, render_template).to_list()

    # If a review app, a canary app or the canonical app while updating config,
    # bust the deployment cache and populate deployments from the yaml file
    if filename or update_config:
      self.cache = []
      for item in yaml_spec:
        if item['kind'] == 'Deployment':
          self.cache.append(item)

    # If updating config, patch the spec and apply
    if update_config:
      print_green("Patching Deployments in spec %s with image digest %s" % (yaml_template, digest), newline_after=True)
      payload = []
      for item in yaml_spec:
        if item['kind'] == 'Deployment':
          item['spec']['progressDeadlineSeconds'] = timeout
          item['metadata']['labels']['app.kubernetes.io/version'] = digest_half
          item['spec']['template']['metadata']['labels']['app.kubernetes.io/version'] = digest_half
          for container in item['spec']['template']['spec']['containers']:
            if self.ecr.project_repo in container['image']:
              container['image'] = "%s@%s" % (self.ecr.project_repo, digest)
        payload.append(item)

      payload_string = yaml.safe_dump_all(
        payload,
        default_flow_style=False
      )
      path = write_temp_file(
        yaml_content_with_header(payload_string),
        HOKUSAI_TMP_DIR
      )
      print_green(
        f'Applying patched spec {path}...', newline_after=True
      )
      self.kctl.apply(path)

    # If not updating config, patch the deployments in the cache and call kubectl patch to update
    else:
      for deployment in self.cache:
        containers = [(container['name'], container['image']) for container in deployment['spec']['template']['spec']['containers']]
        deployment_targets = [
          {
            'name': name,
            'image': "%s@%s" % (self.ecr.project_repo, digest),
          }
          for name, image in containers if self.ecr.project_repo in image
        ]
        patch = {
          "metadata": {
            "labels": {
              "app.kubernetes.io/version": digest_half
            }
          },
          "spec": {
            "template": {
              "metadata": {
                "labels": {
                  "app.kubernetes.io/version": digest_half
                }
              },
              "spec": {
                "containers": deployment_targets
              }
            },
            "progressDeadlineSeconds": timeout
          }
        }

        print_green("Patching deployment %s..." % deployment['metadata']['name'], newline_after=True)
        shout(self.kctl.command("patch deployment %s -p '%s'" % (deployment['metadata']['name'], json.dumps(patch))))

    # Watch the rollouts in the cache and if any fail, roll back
    print_green("Waiting for deployment rollouts to complete...")
    rollout_commands = [self.kctl.command("rollout status deployment/%s" % deployment['metadata']['name']) for deployment in self.cache]
    return_codes = shout_concurrent(rollout_commands, print_output=True)
    if any(return_codes):
      print_red("One or more deployment rollouts failed!  Rolling back...", newline_before=True, newline_after=True)
      rollback_commands = [self.kctl.command("rollout undo deployment/%s" % deployment['metadata']['name']) for deployment in self.cache]
      shout_concurrent(rollback_commands, print_output=True)
      raise HokusaiError("Deployment failed!")

    post_deploy_success = True

    # Run the post-deploy hook for the canonical app or a review app
    if config.post_deploy and (filename is None or (filename and self.namespace)):
      print_green("Running post-deploy hook '%s'..." % config.post_deploy, newline_after=True)
      return_code = CommandRunner(self.context, namespace=self.namespace).run(digest, config.post_deploy, constraint=constraint, tty=False)
      if return_code:
        print_yellow("WARNING: Running the post-deploy hook failed with return code %s" % return_code, newline_before=True, newline_after=True)
        print_yellow("The image digest %s has been rolled out.  However, you should run the post-deploy hook '%s' manually, or re-run this deployment." % (digest, config.post_deploy), newline_after=True)
        post_deploy_success = False

    # For the canonical app, create tags
    if filename is None:
      deployment_tag = "%s--%s" % (self.context, datetime.datetime.utcnow().strftime("%Y-%m-%d--%H-%M-%S"))
      print_green("Updating ECR deployment tags in %s..." % self.ecr.project_repo, newline_after=True)
      try:
        self.ecr.retag(tag, self.context)
        print_green("Updated ECR tag %s -> %s" % (tag, self.context))

        self.ecr.retag(tag, deployment_tag)
        print_green("Updated ECR tag %s -> %s" % (tag, deployment_tag), newline_after=True)
      except (ValueError, ClientError) as e:
        print_yellow("WARNING: Updating ECR deployment tags failed due to the error: '%s'" % str(e), newline_before=True, newline_after=True)
        print_yellow("The tag %s has been rolled out.  However, you should create the ECR tags '%s' and '%s' manually, or re-run this deployment." % (tag, deployment_tag, self.context), newline_after=True)
        post_deploy_success = False

      remote = git_remote or config.git_remote
      if remote:
        # Update git tags. Try up to 3 times, at 3 second intervals. Failure does not fail deployment.
        git_tag_sucess = False
        attempts = 0
        while ((not git_tag_sucess) and (attempts < 3)):
          try:
            attempts += 1
            print_green("Creating Git deployment tags '%s', '%s', and pushing them to %s..." % (self.context, deployment_tag, remote))
            print_green("Attempt# %s." % attempts)
            shout("git fetch -f %s --tags" % remote)
            shout("git tag -f %s %s" % (self.context, tag), print_output=True)
            shout("git tag -f %s %s" % (deployment_tag, tag), print_output=True)
            shout("git push -f --no-verify %s refs/tags/%s" % (remote, self.context), print_output=True)
            print_green("Updated Git tag %s -> %s" % (tag, self.context))
            shout("git push -f --no-verify %s refs/tags/%s" % (remote, deployment_tag), print_output=True)
            print_green("Updated Git tag %s -> %s" % (tag, deployment_tag), newline_after=True)
            git_tag_sucess = True
          except CalledProcessError as e:
            # If subprocess.check_output() was called, the actual error is in CalledProcessError's 'output' attribute.
            print_yellow("WARNING: Creating Git deployment tags failed due to the error:", newline_before=True, newline_after=True)
            print_yellow(e.output)
            time.sleep(3)

        if (not git_tag_sucess):
            print_yellow("Failed all attempts at pushing Git deployment tags! Please do it manually.", newline_after=True)

    if post_deploy_success:
      print_green("Deployment succeeded!")
    else:
      raise HokusaiError("One or more post-deploy steps failed!")

  def refresh(self):
    for deployment in self.cache:
      print_green("Refreshing %s..." % deployment['metadata']['name'], newline_after=True)
      shout(self.kctl.command("rollout restart deployment/%s" % deployment['metadata']['name']))

    print_green("Waiting for refresh to complete...")

    rollout_commands = [self.kctl.command("rollout status deployment/%s" % deployment['metadata']['name']) for deployment in self.cache]
    return_codes = shout_concurrent(rollout_commands, print_output=True)
    if any(return_codes):
      raise HokusaiError("Refresh failed!")


  @property
  def names(self):
    return [deployment['metadata']['name'] for deployment in self.cache]



import os

import yaml

from hokusai import CWD
from hokusai.commands.kubernetes import k8s_delete
from hokusai.lib.common import (
  print_green,
  clean_string,
  local_to_local,
  unlink_file_if_not_debug,
  write_temp_file,
  yaml_content_with_header
)
from hokusai.lib.config import HOKUSAI_CONFIG_DIR, HOKUSAI_TMP_DIR, config
from hokusai.services.configmap import ConfigMap
from hokusai.services.kubectl import Kubectl
from hokusai.services.namespace import Namespace
from hokusai.services.service_account import ServiceAccount
from hokusai.services.yaml_spec import YamlSpec


def copy_configmap(name, destination_namespace):
  ''' copy configmap from default namespace to destination namespace '''
  source_configmap = ConfigMap('staging', name=name)
  destination_configmap = ConfigMap(
    'staging',
    name=name,
    namespace=destination_namespace
  )
  source_configmap.load()
  destination_configmap.struct['data'] = source_configmap.struct['data']
  destination_configmap.save()

def copy_service_account(name, destination_namespace):
  ''' copy service account from default namespace to destination namespace '''
  source_sa = ServiceAccount('staging', name=name)
  source_sa.load()
  spec = source_sa.spec
  spec['metadata']['namespace'] = destination_namespace
  dest_sa = ServiceAccount(
    'staging',
    namespace=destination_namespace,
    name=name,
    spec=spec
  )
  dest_sa.apply()

def create_yaml(source_file, app_name):
  ''' create yaml for review app '''
  yaml_content = YamlSpec(source_file).to_list()
  namespace = clean_string(app_name)
  for k8s_resource in yaml_content:
    edit_namespace(k8s_resource, namespace)
  payload_string = yaml.safe_dump_all(
    yaml_content,
    default_flow_style=False
  )
  tmp_path = write_temp_file(
    yaml_content_with_header(payload_string),
    HOKUSAI_TMP_DIR
  )
  local_to_local(
    tmp_path,
    os.path.join(CWD, HOKUSAI_CONFIG_DIR),
    f'{app_name}.yml',
    create_target_dir=False
  )
  unlink_file_if_not_debug(tmp_path)
  path = f'{HOKUSAI_CONFIG_DIR}/{app_name}.yml'
  print_green(f'Created {path}')
  # this return causes Hokusai to exit with 1
  # so this function must be wrapped if called from CLI
  return path

def create_yaml_for_cli(source_file, app_name):
  ''' wrap create_yaml function '''
  create_yaml(source_file, app_name)

def delete_review_app(context, app_name, filename):
  ''' delete review app '''
  namespace = clean_string(app_name)
  # delete resources in yaml
  k8s_delete(context, namespace, filename)
  # delete namespace
  ns = Namespace('staging', namespace)
  ns.delete()

def edit_namespace(k8s_resource, destination_namespace):
  ''' edit namespace field for a Kubernetes resource definition '''
  if 'apiVersion' in k8s_resource:
    struct = {
      'namespace': destination_namespace
    }
    k8s_resource['metadata'] = k8s_resource.setdefault(
      'metadata', struct
    ) | struct

def list_review_apps(context, labels=None):
  ''' list project's review apps '''
  kctl = Kubectl(context)
  namespaces = kctl.get_objects('namespaces', labels)
  for ns in namespaces:
    print(ns['metadata']['name'])

def setup_review_app(source_file, app_name):
  ''' prepare for creating a review app '''
  namespace = clean_string(app_name)

  # create namespace
  labels = {
    'app-name': config.project_name,
    'app-phase': 'review'
  }
  ns = Namespace('staging', namespace, labels)
  ns.create()

  # create yaml file
  path = create_yaml(source_file, app_name)

  # get list of configmaps referenced in yaml's Deployments
  # and copy them to review app namespace
  configmap_refs = YamlSpec(
    path,
    render_template=False
  ).all_deployments_configmap_refs()
  for configmap in configmap_refs:
    print_green(
      f'Copying {configmap} ConfigMap to {namespace} namespace...'
    )
    copy_configmap(configmap, namespace)

  # get list of service accounts referenced in yaml's Deployments
  # and copy them to review app namespace
  service_account_refs = YamlSpec(
    path,
    render_template=False
  ).all_deployments_sa_refs()
  for sa in service_account_refs:
    print_green(
      f'Copying {sa} ServiceAccount to {namespace} namespace...'
    )
    copy_service_account(sa, namespace)



import os

import json
import yaml

from hokusai.lib.common import shout, unlink_file_if_not_debug
from hokusai.lib.global_config import HokusaiGlobalConfig


class Kubectl:
  ''' run kubectl commands '''
  def __init__(self, context, namespace=None):
    self.context = context
    self.namespace = namespace
    global_config = HokusaiGlobalConfig()
    self.kubectl = os.path.join(
      global_config.kubectl_dir, 'kubectl'
    )

  def _apply_or_create(self, action, k8s_spec_file):
    ''' run kubectl apply or create on a k8s spec file '''
    try:
      shout(
        self.command(f'{action} -f {k8s_spec_file}'),
        print_output=True
      )
    finally:
      unlink_file_if_not_debug(k8s_spec_file)

  def apply(self, k8s_spec_file):
    ''' run kubectl apply on a k8s spec file '''
    self._apply_or_create('apply', k8s_spec_file)

  def command(self, cmd):
    ''' generate kubectl command '''
    if self.namespace is None:
      return f'{self.kubectl} --context {self.context} {cmd}'
    return (
      f'{self.kubectl} --context {self.context} ' +
      f'--namespace {self.namespace} {cmd}'
    )

  def contexts(self):
    ''' return all k8s contexts found in kubeconfig '''
    return [
      context['name'] for context in
      yaml.load(
        shout(f'{self.kubectl} config view'),
        Loader=yaml.FullLoader
      )['contexts']
    ]

  def create(self, k8s_spec_file):
    ''' run kubectl create on a k8s spec file '''
    self._apply_or_create('create', k8s_spec_file)

  def get_object(self, obj):
    ''' run kubectl get <object> '''
    cmd = self.command(f'get {obj} -o json')
    try:
      return json.loads(shout(cmd))
    except ValueError:
      return None

  def get_objects(self, obj, selector=None):
    ''' run kubectl get <object> with selectors '''
    if selector is not None:
      cmd = self.command(
        f'get {obj} --selector {selector} -o json'
      )
    else:
      cmd = self.command(f'get {obj} -o json')
    try:
      return json.loads(shout(cmd))['items']
    except ValueError:
      return []



import os

from hokusai.lib.common import shout
from hokusai.lib.config import BUILD_YAML_FILE, config
from hokusai.lib.docker_compose_helpers import generate_compose_command


class Docker:
  def build(self, filename=None):
    env_vars = "DOCKER_DEFAULT_PLATFORM=linux/amd64"
    compose_command = generate_compose_command(
      filename,
      default_yaml_file=BUILD_YAML_FILE
    )
    opts = "--progress plain"
    build_command = f'{env_vars} {compose_command} -p hokusai build {opts}'
    if config.pre_build:
      build_command = "%s && %s" % (config.pre_build, build_command)
    if config.post_build:
      build_command = "%s && %s" % (build_command, config.post_build)
    shout(build_command, print_output=True)



import os

from hokusai import CWD
from hokusai.lib.common import print_green, shout, returncode
from hokusai.lib.config import HOKUSAI_CONFIG_DIR, config
from hokusai.lib.exceptions import HokusaiError
from hokusai.lib.template_selector import TemplateSelector
from hokusai.services.configmap import ConfigMap
from hokusai.services.ecr import ECR
from hokusai.services.kubectl import Kubectl
from hokusai.services.yaml_spec import YamlSpec


def k8s_create(
  context,
  tag='latest',
  namespace=None,
  filename=None,
  environment=(),
  render_template=True
):
  if filename is None:
    yaml_template = TemplateSelector().get(os.path.join(CWD, HOKUSAI_CONFIG_DIR, context))
  else:
    yaml_template = TemplateSelector().get(filename)

  ecr = ECR()
  if not ecr.project_repo_exists():
    raise HokusaiError("ECR repository %s does not exist... did you run `hokusai setup` for this project?" % config.project_name)

  if not ecr.tag_exists(tag):
    raise HokusaiError("Image tag %s does not exist... did you run `hokusai registry push`?" % tag)

  if tag == 'latest' and not ecr.tag_exists(context):
    ecr.retag(tag, context)
    print_green("Updated tag 'latest' -> %s" % context)

  if filename is None:
    configmap = ConfigMap(context, namespace=namespace)
    for s in environment:
      if '=' not in s:
        raise HokusaiError("Error: environment variables must be of the form 'KEY=VALUE'")
      split = s.split('=', 1)
      configmap.update(split[0], split[1])
    configmap.create()
    print_green("Created configmap %s-environment" % config.project_name)

  kctl = Kubectl(context, namespace=namespace)
  yaml_spec = YamlSpec(yaml_template, render_template).to_file()

  shout(kctl.command("create --save-config -f %s" % yaml_spec), print_output=True)
  print_green("Created Kubernetes environment %s" % yaml_template)


def k8s_update(
  context,
  namespace=None,
  filename=None,
  check_branch="main",
  check_remote=None,
  skip_checks=False,
  dry_run=False,
  render_template=True
):
  if filename is None:
    yaml_template = TemplateSelector().get(os.path.join(CWD, HOKUSAI_CONFIG_DIR, context))
  else:
    yaml_template = TemplateSelector().get(filename)

  if not skip_checks:
    current_branch = None
    for branchname in shout('git branch').splitlines():
      if '* ' in branchname:
        current_branch = branchname.replace('* ', '')
        break

    if 'detached' in current_branch:
      raise HokusaiError("Not on any branch!  Aborting.")
    if current_branch != check_branch:
      raise HokusaiError("Not on %s branch!  Aborting." % check_branch)

    remotes = [check_remote] if check_remote else shout('git remote').splitlines()
    for remote in remotes:
      shout("git fetch %s" % remote)
      if returncode("git diff --quiet %s/%s" % (remote, current_branch)):
        raise HokusaiError("Local branch %s is divergent from %s/%s.  Aborting." % (current_branch, remote, current_branch))

  kctl = Kubectl(context, namespace=namespace)
  yaml_spec = YamlSpec(yaml_template, render_template).to_file()

  if dry_run:
    shout(kctl.command("apply -f %s --dry-run" % yaml_spec), print_output=True)
    print_green("Updated Kubernetes environment %s (dry run)" % yaml_template)
  else:
    shout(kctl.command("apply -f %s" % yaml_spec), print_output=True)
    print_green("Updated Kubernetes environment %s" % yaml_template)


def k8s_delete(context, namespace=None, filename=None, render_template=True):
  if filename is None:
    yaml_template = TemplateSelector().get(os.path.join(CWD, HOKUSAI_CONFIG_DIR, context))
  else:
    yaml_template = TemplateSelector().get(filename)

  if filename is None:
    configmap = ConfigMap(context, namespace=namespace)
    configmap.destroy()
    print_green("Deleted configmap %s-environment" % config.project_name)

  kctl = Kubectl(context, namespace=namespace)
  yaml_spec = YamlSpec(yaml_template, render_template).to_file()

  shout(kctl.command("delete -f %s" % yaml_spec), print_output=True)
  print_green("Deleted Kubernetes environment %s" % yaml_template)


def k8s_status(
  context,
  resources,
  pods,
  describe,
  top,
  namespace=None,
  filename=None,
  render_template=True
):
  if filename is None:
    yaml_template = TemplateSelector().get(os.path.join(CWD, HOKUSAI_CONFIG_DIR, context))
  else:
    yaml_template = TemplateSelector().get(filename)

  kctl = Kubectl(context, namespace=namespace)
  yaml_spec = YamlSpec(yaml_template, render_template).to_file()

  if describe:
    kctl_cmd = "describe"
    output = ""
  else:
    kctl_cmd = "get"
    output = " -o wide"
  if resources:
    print_green("Resources", newline_before=True)
    print_green("===========")
    shout(kctl.command("%s -f %s%s" % (kctl_cmd, yaml_spec, output)), print_output=True)
  if pods:
    print_green("Pods", newline_before=True)
    print_green("===========")
    shout(kctl.command("%s pods --selector app=%s,layer=application%s" % (kctl_cmd, config.project_name, output)), print_output=True)
  if top:
    print_green("Top Pods", newline_before=True)
    print_green("===========")
    shout(kctl.command("top pods --selector app=%s,layer=application" % config.project_name), print_output=True)



import textwrap

import click

class CommandTreeNode():
  """Represent a node in a tree of Click commands"""
  def __init__(self, click_command):
    self.click_command = click_command
    self.children = []

def build_tree(click_command):
  """Build a tree of CommandTreeNode starting from the given click_command"""
  node = CommandTreeNode(click_command)
  if isinstance(click_command, click.core.Group):
    node.children = [build_tree(cmd) for _, cmd in sorted(click_command.commands.items())]
  return node

def print_command_in_tree(click_command, last=False, prefix=''):
  """Print the name and docstring of the given Click command, as part of a tree output"""
  if last:
    hanger = '└──'
  else:
    hanger = '├──'
  line = prefix + hanger + " " + click_command.name
  if click_command.help:
    line += " - " + textwrap.shorten(click_command.help, 100, placeholder='...')
  print(line)

def print_command_tree(click_command):
  """Print all sub-commands of the given click_command, as a tree"""
  root = build_tree(click_command)
  print_tree(root, level=1)

def print_tree(start_node, level, prefix=''):
  """Print CommandTreeNode tree from the given start_node"""
  num_children = len(start_node.children)
  for i, node in enumerate(start_node.children):
    if i == num_children - 1:
      last = True
      next_level_prefix = prefix + '    '
    else:
      last = False
      next_level_prefix = prefix + '│   '
    print_command_in_tree(node.click_command, last=last, prefix=prefix)
    print_tree(node, level=level+1, prefix=next_level_prefix)



from hokusai.lib.common import print_green
from hokusai.services.deployment import Deployment
from hokusai.services.command_runner import CommandRunner
from hokusai.services.ecr import ECR
from hokusai.lib.exceptions import HokusaiError

def update(
  context,
  tag,
  migration,
  constraint,
  git_remote,
  timeout,
  namespace=None,
  update_config=False,
  filename=None,
  render_template=True
):
  if migration is not None:
    print_green("Running migration '%s' on %s..." % (migration, context), newline_after=True)
    return_code = CommandRunner(context, namespace=namespace).run(tag, migration, constraint=constraint, tty=False)
    if return_code:
      raise HokusaiError("Migration failed with return code %s" % return_code, return_code=return_code)
  Deployment(context, namespace=namespace).update(
    tag,
    constraint,
    git_remote,
    timeout,
    update_config=update_config,
    filename=filename,
    render_template=render_template
  )
  print_green("Deployment(s) updated to %s" % tag)


def refresh(context, deployment_name, namespace=None):
  deployment = Deployment(context, deployment_name=deployment_name, namespace=namespace)
  deployment.refresh()


def promote(migration, constraint, git_remote, timeout, update_config=False, filename=None):
  if migration is not None:
    print_green("Running migration '%s' on production..." % migration, newline_after=True)
    return_code = CommandRunner('production').run('staging', migration, constraint=constraint, tty=False)
    if return_code:
      raise HokusaiError("Migration failed with return code %s" % return_code, return_code=return_code)

  Deployment('production').update('staging', constraint, git_remote, timeout, update_config=update_config, filename=filename)
  print_green("Promoted staging to production")



from hokusai.lib.config import config
from hokusai.services.configmap import ConfigMap
from hokusai.lib.common import print_green, print_smart
from hokusai.lib.exceptions import HokusaiError


def delete_env(context, namespace=None):
  configmap = ConfigMap(context, namespace=namespace)
  configmap.destroy()
  print_green("Deleted configmap %s-environment" % config.project_name)


def get_env(context, environment, namespace=None):
  configmap = ConfigMap(context, namespace=namespace)
  configmap.load()
  for k, v in sorted(configmap.all().items()):
    if not environment or (environment and k in environment):
      print_smart("%s=%s" % (k, v))

def set_env(context, environment, namespace=None):
  configmap = ConfigMap(context, namespace=namespace)
  configmap.load()
  for s in environment:
    if '=' not in s:
      raise HokusaiError("Error: environment variables must be of the form 'KEY=VALUE'")
    split = s.split('=', 1)
    configmap.update(split[0], split[1])
  configmap.save()


def unset_env(context, environment, namespace=None):
  configmap = ConfigMap(context, namespace=namespace)
  configmap.load()
  for s in environment:
    configmap.delete(s)
  configmap.save()



import copy
import json

from hokusai.lib.common import delete_keys, shout, write_temp_file
from hokusai.lib.config import HOKUSAI_TMP_DIR
from hokusai.services.kubectl import Kubectl


class ServiceAccount:
  ''' represent a Kubernetes ServiceAccount '''

  k8s_dynamic_fields = {
    'metadata': {
      'annotations': {
        'kubectl.kubernetes.io/last-applied-configuration': {}
      },
      'creationTimestamp': {},
      'resourceVersion': {},
      'uid': {}
    },
    'secrets': {}
  }

  def __init__(
      self,
      context,
      namespace='default',
      name=None,
      spec={}

  ):
    self.context = context
    self.kctl = Kubectl(context, namespace=namespace)
    self.name = name

    # if resource exists in Kubernetes,
    # its representation should be stored here.
    # should be in sync with self.spec
    self.object = {}

    # store the spec that can be used
    # to create/update the object in Kubernetes.
    # should be in sync with self.object,
    # if object exists in Kubernetes
    self.spec = spec

  def apply(self):
    ''' apply service account spec to Kubernetes '''
    path = write_temp_file(
      json.dumps(self.spec),
      HOKUSAI_TMP_DIR
    )
    self.kctl.apply(path)
    # should call self.load to update object,
    # but will have to account for delay in Kubernetes apply

  def load(self):
    ''' read object from Kubernetes and store it '''
    payload = shout(
      self.kctl.command(
        f'get serviceaccount {self.name} -o json'
      )
    )
    self.object = json.loads(payload)
    # keep spec in sync
    spec = copy.deepcopy(self.object)
    delete_keys(
      spec,
      self.k8s_dynamic_fields
    )
    self.spec = spec



import boto3
import botocore.exceptions as botoexceptions
import os

from hokusai import CWD
from hokusai.lib.common import get_region_name, print_green, print_red, shout
from hokusai.lib.config import HOKUSAI_CONFIG_DIR, BUILD_YAML_FILE, TEST_YML_FILE, DEVELOPMENT_YML_FILE, config
from hokusai.lib.docker_compose_helpers import detect_compose_command
from hokusai.lib.exceptions import CalledProcessError, HokusaiError
from hokusai.lib.template_selector import TemplateSelector
from hokusai.services.ecr import ECR
from hokusai.services.kubectl import Kubectl

def check():
  return_code = 0

  def check_ok(check_item):
    print_green('\u2714 ' + check_item + ' found')

  def check_err(check_item):
    print_red('\u2718 ' + check_item + ' not found')

  try:
    config.project_name
    check_ok('Config project-name')
  except HokusaiError:
    check_err('Config project-name')

  try:
    shout('which docker')
    check_ok('docker')
  except CalledProcessError:
    check_err('docker')
    return_code += 1

  compose_command = detect_compose_command()
  print_green(
    '\u2714 ' +
    f'Will use "{compose_command}" command for Docker Compose'
  )

  try:
    shout('which kubectl')
    check_ok('kubectl')
  except CalledProcessError:
    check_err('kubectl')
    return_code += 1

  try:
    shout('which git')
    check_ok('git')
  except CalledProcessError:
    check_err('git')
    return_code += 1

  try:
    boto3.client('sts', region_name=get_region_name()).get_caller_identity()
    check_ok('Valid AWS credentials')
  except (botoexceptions.ClientError, botoexceptions.NoCredentialsError):
    check_err('Valid AWS credentials')
    return_code += 1

  ecr = ECR()
  if ecr.project_repo_exists():
    check_ok("ECR repository '%s'" % config.project_name)
  else:
    check_err("ECR repository '%s'" % config.project_name)
    return_code += 1

  try:
    build_template = TemplateSelector().get(os.path.join(CWD, HOKUSAI_CONFIG_DIR, BUILD_YAML_FILE))
    check_ok(HOKUSAI_CONFIG_DIR + '/' + os.path.split(build_template)[-1])
  except HokusaiError:
    check_err('hokusai/build.*')
    return_code += 1

  try:
    development_template = TemplateSelector().get(os.path.join(CWD, HOKUSAI_CONFIG_DIR, DEVELOPMENT_YML_FILE))
    check_ok(HOKUSAI_CONFIG_DIR + '/' + os.path.split(development_template)[-1])
  except HokusaiError:
    check_err('hokusai/development.*')
    return_code += 1

  try:
    test_template = TemplateSelector().get(os.path.join(CWD, HOKUSAI_CONFIG_DIR, TEST_YML_FILE))
    check_ok(HOKUSAI_CONFIG_DIR + '/' + os.path.split(test_template)[-1])
  except HokusaiError:
    check_err('hokusai/test.*')
    return_code += 1

  for context in ['staging', 'production']:
    try:
      if context in Kubectl('staging').contexts():
        check_ok("kubectl context '%s'" % context)
      else:
        check_err("kubectl context '%s'" % context)
        return_code += 1
    except CalledProcessError:
      check_err('%s context' % context)
      return_code += 1

    try:
      context_template = TemplateSelector().get(os.path.join(CWD, HOKUSAI_CONFIG_DIR, context))
      check_ok(HOKUSAI_CONFIG_DIR + '/' + os.path.split(context_template)[-1])
    except HokusaiError:
      check_err("hokusai/%s.*" % context)
      return_code += 1

  return return_code



from hokusai.lib.common import print_green, print_red
from hokusai.lib.exceptions import HokusaiError
from hokusai.services.ecr import ECR, ClientError

def retag(tag_to_change, tag_to_match):
  ecr = ECR()

  if not ecr.project_repo_exists():
    raise HokusaiError("Project repo does not exist. Aborting.")

  try:
    ecr.retag(tag_to_match, tag_to_change)
    print_green("Updated ECR '%s' tag to point to the image that '%s' tag points to." % (tag_to_change, tag_to_match), newline_after=True)
  except (ValueError, ClientError) as e:
    raise HokusaiError("Updating ECR tag failed due to the error: '%s'" % str(e))



import os

from hokusai.lib.config import config
from hokusai.services.ecr import ECR
from hokusai.lib.common import print_green, shout
from hokusai.lib.exceptions import HokusaiError

def pull(tag, local_tag):
  ecr = ECR()
  if not ecr.project_repo_exists():
    raise HokusaiError("ECR repo %s does not exist... did you run `hokusai setup` for this project?" % config.project_name)

  shout(ecr.get_login(), mask=(r'^(docker login -u) .+ (-p) .+ (.+)$', r'\1 ****** \2 ***** \3'))

  shout("docker pull %s:%s" % (ecr.project_repo, tag))

  shout("docker tag %s:%s hokusai_%s:%s" % (ecr.project_repo, tag, config.project_name, local_tag))
  
  print_green("Pulled %s:%s to hokusai_%s:%s" % (ecr.project_repo, tag, config.project_name, local_tag), newline_after=True)



from hokusai.lib.common import print_green, shout
from hokusai.services.deployment import Deployment
from hokusai.services.ecr import ECR
from hokusai.lib.exceptions import HokusaiError

def gitdiff():
  ecr = ECR()

  staging_tag = ecr.find_git_sha1_image_tag('staging')
  if staging_tag is None:
    raise HokusaiError("Could not find a tag for staging.  Aborting.")

  production_tag = ecr.find_git_sha1_image_tag('production')
  if production_tag is None:
    raise HokusaiError("Could not find a git SHA1 tag for production.  Aborting.")

  print_green("Comparing %s to %s" % (production_tag, staging_tag))
  for remote in shout('git remote').splitlines():
    shout("git fetch %s" % remote)
  shout("git diff %s %s" % (production_tag, staging_tag), print_output=True)



from hokusai.commands.build import build
from hokusai.commands.check import check
from hokusai.commands.command_tree import print_command_tree
from hokusai.commands.configure import hokusai_configure
from hokusai.commands.deployment import update, refresh, promote
from hokusai.commands.development import dev_start, dev_stop, dev_status, dev_logs, dev_run, dev_clean
from hokusai.commands.gitdiff import gitdiff
from hokusai.commands.gitlog import gitlog
from hokusai.commands.gitcompare import gitcompare
from hokusai.commands.env import get_env, set_env, unset_env
from hokusai.commands.images import images
from hokusai.commands.logs import logs
from hokusai.commands.push import push_image
from hokusai.commands.pull import pull
from hokusai.commands.run import run
from hokusai.commands.setup import setup
from hokusai.commands.kubernetes import k8s_create, k8s_update, k8s_delete, k8s_status
from hokusai.commands.review_app import create_yaml_for_cli, delete_review_app, list_review_apps, setup_review_app
from hokusai.commands.retag import retag
from hokusai.commands.test import test
from hokusai.commands.version import version



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['hokusai',
 'hokusai.cli',
 'hokusai.commands',
 'hokusai.lib',
 'hokusai.services']

package_data = \
{'': ['*'], 'hokusai': ['templates/*', 'templates/hokusai/*']}

install_requires = \
['Jinja2==3.1.1',
 'MarkupSafe==2.1.1',
 'PyYAML==6.0.1',
 'boto3>=1.34.25,<2.0.0',
 'click-repl==0.2.0',
 'click==8.1.0',
 'packaging>=24.1,<25.0',
 'prompt-toolkit==3.0.28',
 'termcolor==1.1.0']

entry_points = \
{'console_scripts': ['hokusai = hokusai.command_line:main']}

setup_kwargs = {
    'name': 'hokusai',
    'version': '3.3.0',
    'description': 'Artsy Docker development toolkit',
    'long_description': '## HOKUSAI [![CircleCI](https://circleci.com/gh/artsy/hokusai/tree/main.svg?style=svg)](https://circleci.com/gh/artsy/hokusai/tree/main)\n\n<a href="https://en.wikipedia.org/wiki/Hokusai"><img height="300" src="hokusai.jpg"></a>\n\nHokusai is a Docker + Kubernetes CLI for application developers.\n\nHokusai "dockerizes" applications and manages their lifecycle throughout development, testing, and release cycles.\n\nHokusai wraps calls to [Kubectl](https://kubernetes.io/), [Docker](https://www.docker.com/), [Docker-Compose](https://docs.docker.com/compose/) and [Git](https://git-scm.com/) with a CLI, and defines a CI workflow.\n\nHokusai currently only supports Kubernetes deployments on AWS, configured to pull from ECS container repositories (ECR), although other providers may be added in the future.\n\n### Why Hokusai?\n\nAt [Artsy](http://www.artsy.net), as we began working with Kubernetes, while impressed with its design, capabilities, and flexibility, we were in need of tooling we could deliver to agile development teams that addressed the day-to-day tasks of application development, delivery, introspection and maintenance, while providing a clean and uncomplicated interface.\n\nTransitioning teams to the Docker / Kubernetes ecosystem can be intimidating, and comes with a steep learning curve. We set out to create a Heroku-like CLI that would shepherd the application developer into the ecosystems of Docker and Kubernetes, and while introducing new tooling and concepts, outlining a clear practice for dependency management, local development, testing and CI, image repository structure, deployment and orchestration.\n\n### Point People\n- [@artsyjian](https://github.com/artsyjian)\n\n## Installation\n\n### MacOS\n\nWe recommend installing via Homebrew:\n\n```\n$ brew update\n$ brew tap artsy/formulas\n$ brew install hokusai\n```\n\nIf you previously installed Hokusai via an alternate installation method, you may need to force the `link` step.\n\n```\n$ brew link --overwrite hokusai\n```\n\nIf you previously installed Hokusai via Pip, you may want to first uninstall it:\n\n```\n$ pip uninstall hokusai\n```\n\n### Linux\n\n```\ncurl -sSL https://raw.githubusercontent.com/artsy/hokusai/main/get-hokusai.sh | sudo bash\n```\n\nNote: This method installs Hokusai to `/usr/local/bin/hokusai`.\n\n### Pip\n\nHokusai can be installed via Pip, on MacOS or Linux. If you do so, please first go through [Pyenv](#Pyenv), [Python](#Python), and [Virtualenv](#Virtualenv) steps.\n\nPython 3.7+ is required.\n\n```\npip install hokusai\n```\n\nNote: If Pip fails at upgrading your system Python packages, try:\n\n```\npip install hokusai --ignore-installed\n```\n\n### Docker\n\nWe also maintain [Hokusai Docker images](https://hub.docker.com/r/artsy/hokusai) for running Hokusai in Docker.\n\n### Github\n\nRelease artifacts are available on [Github](https://github.com/artsy/hokusai/releases).\n\n### AWS S3\n\nRelease artifacts are also available in AWS S3. You can use this [convenience script](get-hokusai.sh) or Curl to fetch them.\n\n### A note on Python 2.x\n\nHokusai currently supports Python 3.7+ only. The last version that supported Python 2.x was [v0.5.18](https://github.com/artsy/hokusai/tree/v0.5.18).\n\n## Setup\n\nWe assume that your org admin has already set up a Kubernetes cluster and an AWS infrastructure, and that your local environment has Git, Docker, and Docker-Compose installed. Perform the following steps to get Hokusai working:\n\n1. Ensure your org admin has completed the steps mentioned in [Administering Hokusai](./docs/Administering_Hokusai.md).\n\n2. Configure your local environment with [AWS credentials](https://boto3.amazonaws.com/v1/documentation/api/latest/guide/configuration.html#configuring-credentials).\n\n3. Run [hokusai configure command](docs/Command_Reference.md#configuring-hokusai-for-your-organization).\n\n4. Optionally, enable Bash autocompletion:\n\n    ```\n    eval "$(_HOKUSAI_COMPLETE=source hokusai)"\n    ```\n\n## Getting Started\n\nOnce Hokusai is configured, you can start using it on a project. Please see [Getting Started](./docs/Getting_Started.md).\n\n## Command Reference\n\nA full command reference can be found in [Command Reference](./docs/Command_Reference.md).\n\n## Review Apps\n\nHokusai can be used to simplify the process of spinning up a "review app" instance of your project, based on a feature branch or pull request.\n\nFull details are in the [Review App reference](./docs/Review_Apps.md).\n\n## Hokusai Files\n\nA descripton of all the [files used by Hokusai](docs/hokusai_files.md).\n\n## Developing Hokusai\n\nTo work on Hokusai itself, please set up:\n\n### Pyenv\n\nWe recommend using [Pyenv](https://github.com/pyenv/pyenv) to install the correct version of Python. For a tutorial of Pyenv, see [this guide](https://realpython.com/intro-to-pyenv/).\n\nWhen installing on MacOS, please make sure to use brew-installed `openssl` and `readline` libraries, and xcode-installed `zlib` library. And make sure these libraries are correctly linked. Like so:\n\n```\nbrew install openssl readline zlib\n\necho \'export PATH="/usr/local/opt/openssl@1.1/bin:$PATH"\' >> ~/.bash_profile\necho \'export LDFLAGS="-L/usr/local/opt/openssl@1.1/lib"\' >> ~/.bash_profile\necho \'export CPPFLAGS="-I/usr/local/opt/openssl@1.1/include"\' >> ~/.bash_profile\necho \'export PKG_CONFIG_PATH="/usr/local/opt/openssl@1.1/lib/pkgconfig"\' >> ~/.bash_profile\n```\n\n### Python\n\nHokusai is currently tested on Python 3.10 so we recommend using that Python version.\n\nIf you use Pyenv to install Python, you should see an output similar to this:\n\n```\n$ pyenv install 3.10\npython-build: use openssl@1.1 from homebrew\npython-build: use readline from homebrew\nDownloading Python-3.10.13.tar.xz...\n-> https://www.python.org/ftp/python/3.10.13/Python-3.10.13.tar.xz\nInstalling Python-3.10.13...\npython-build: use tcl-tk from homebrew\npython-build: use readline from homebrew\npython-build: use ncurses from homebrew\npython-build: use zlib from xcode sdk\nInstalled Python-3.10.13 to /Users/jxu/.pyenv/versions/3.10.13\n```\n\nWith the desired Python version installed, activate it globally:\n\n```\npyenv global 3.10\n```\n\nNote: If you want to create a PyInstaller distribution, Python must be installed with development libraries. Use the environment variable `PYTHON_CONFIGURE_OPTS="--enable-framework"` on Darwin and `PYTHON_CONFIGURE_OPTS="--enable-shared"` on Linux when running `pyenv install`.\n\n### Virtualenv\n\nWe recommend using a virtual environment to isolate Hokusai\'s dependencies from that of other projects on your local environment.\n\nThe Pyenv install comes with [pyenv-virtualenv](https://github.com/pyenv/pyenv-virtualenv) which can be used to create virtual environments.\n\n### Poetry\n\nUse [Poetry](https://python-poetry.org/) to install Hokusai\'s dependencies as well as Hokusai itself in [editable mode](https://pip-python3.readthedocs.io/en/latest/reference/pip_install.html#editable-installs). Please see [this guide](https://python-poetry.org/docs/basic-usage/) for working with Poetry.\n\nInstall Poetry:\n\n```\npip install --upgrade pip\npip install poetry\n```\n\nInstall dependencies and Hokusai in editable mode:\n\n```\npoetry install\n```\n\nTo update dependencies:\n\n```\npoetry lock\n```\n\n## Testing\n\n### Install Minikube\n\n[Minikube](https://minikube.sigs.k8s.io/docs/start/) is used for integration tests.\n\n```\nbrew install minikube\nminikube start --kubernetes-version=<version of your Kubernetes clusters, example: v1.2.3>\n```\n\n### Run tests\n\nTo run unit tests:\n\n```\nmake test\n```\n\nTo run integration tests:\n\n```\nmake integration-local\n```\n\nOnly specific modules, TestClasses, or even methods:\n\n```\npython -m unittest test.unit.test_module.TestClass.test_method\n```\n\nTip: Set `DEBUG=1` environment variable to print boto logging\n\n\n## Distributing Hokusai\n\n### Beta Release\n\nMerging a branch into `main` automatically creates a beta version useful for testing. A [script](./scripts/update_version_file.sh) that runs in CI automatically generates a Beta version number and writes it into [VERSION](./hokusai/VERSION) file. The version number is based on the combination of the latest canonical version found in [RELEASE_VERSION](./hokusai/RELEASE_VERSION) file and the latest Git commit in `main` branch. [VERSION](./hokusai/VERSION) file in version control has just a dummy number (e.g. 999.999.999).\n\nTo install the beta:\n\n#### MacOS\n\n```\n$ brew uninstall hokusai\n$ brew uninstall hokusai-beta\n$ brew update\n$ brew tap artsy/formulas\n$ brew install hokusai-beta\n```\n\n#### Linux\n\n```\ncurl -sSL https://raw.githubusercontent.com/artsy/hokusai/main/get-hokusai.sh | sudo bash -s beta\n```\n\n### Official Release\n\nTo create an official release, such as `v1.2.3`, perform the following:\n\n- Create a branch named `prepare-v1.2.3` and make the following changes:\n  - Bump canonical version in [RELEASE_VERSION](./hokusai/RELEASE_VERSION) file.\n  - Upate [CHANGELOG](./CHANGELOG.md).\n  - Open a PR to merge into `main`. Please see [past PRs](https://github.com/artsy/hokusai/pulls?q=is%3Apr+Release+is%3Aclosed+%22prepare+version%22) for example.\n\n- [Open a PR](https://github.com/artsy/hokusai/compare/release...main?expand=1) to merge `main` into `release`. Please see [past PRs](https://github.com/artsy/hokusai/pulls?q=is%3Apr+is%3Aclosed+%22release+version%22) for example.\n\nA CI [script](./scripts/update_version_file.sh) will copy the version in [RELEASE_VERSION](./hokusai/RELEASE_VERSION) file to [VERSION](./hokusai/VERSION) file.\n\n## The Name\n\nThe project is named for the great Japanese artist [Katsushika Hokusai](https://www.artsy.net/article/artsy-editorial-7-things-hokusai-creator-great-wave) (1760-1849).\n\n## About Artsy\n\n<a href="https://www.artsy.net/">\n  <img align="left" src="https://avatars2.githubusercontent.com/u/546231?s=200&v=4"/>\n</a>\n\nThis project is the work of engineers at [Artsy][footer_website], the world\'s\nleading and largest online art marketplace and platform for discovering art.\nOne of our core [Engineering Principles][footer_principles] is being [Open\nSource by Default][footer_open] which means we strive to share as many details\nof our work as possible.\n\nYou can learn more about this work from [our blog][footer_blog] and by following\n[@ArtsyOpenSource][footer_twitter] or explore our public data by checking out\n[our API][footer_api]. If you\'re interested in a career at Artsy, read through\nour [job postings][footer_jobs]!\n\n[footer_website]: https://www.artsy.net/\n[footer_principles]: https://github.com/artsy/README/blob/main/culture/engineering-principles.md\n[footer_open]: https://github.com/artsy/README/blob/main/culture/engineering-principles.md\n[footer_blog]: https://artsy.github.io/\n[footer_twitter]: https://twitter.com/ArtsyOpenSource\n[footer_api]: https://developers.artsy.net/\n[footer_jobs]: https://www.artsy.net/jobs\n',
    'author': 'Isac Petruzzi',
    'author_email': 'isac@artsymail.com',
    'maintainer': 'Jian-Feng Xu',
    'maintainer_email': 'jian-feng@artsymail.com',
    'url': 'http://github.com/artsy/hokusai',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<3.11',
}


setup(**setup_kwargs)



import click

import hokusai

from hokusai.cli.base import base
from hokusai.lib.command import command
from hokusai.lib.common import set_verbosity, CONTEXT_SETTINGS

@base.group()
def pipeline(context_settings=CONTEXT_SETTINGS):
  """Interact with the project's' staging -> production pipeline"""
  pass


@pipeline.command(context_settings=CONTEXT_SETTINGS)
@click.option('-v', '--verbose', type=click.BOOL, is_flag=True, help='Verbose output')
def gitdiff(verbose):
  """Print a git diff between the tag currently deployed on production
  and the tag currently deployed on staging"""
  set_verbosity(verbose)
  command(
    hokusai.gitdiff
  )


@pipeline.command(context_settings=CONTEXT_SETTINGS)
@click.option('-v', '--verbose', type=click.BOOL, is_flag=True, help='Verbose output')
def gitlog(verbose):
  """Print a git log between the tag currently deployed on production
  and the tag currently deployed on staging"""
  set_verbosity(verbose)
  command(
    hokusai.gitlog
  )

@pipeline.command(context_settings=CONTEXT_SETTINGS)
@click.option('--org-name', type=click.STRING, required=True, help='Name of the (github) organization')
@click.option('--git-compare-link', type=click.STRING, help='Python formatted string input that gets org name, project name and 2 diff sha1s, example: https://github.com/%s/%s/compare/%s...%s', default="https://github.com/%s/%s/compare/%s...%s")
@click.option('-v', '--verbose', type=click.BOOL, is_flag=True, help='Verbose output')
def gitcompare(org_name, git_compare_link, verbose):
  """Prints a git compare link between the tag currently deployed on production
  and the tag currently deployed on staging"""
  set_verbosity(verbose)
  command(
    hokusai.gitcompare,
    org_name,
    git_compare_link
  )


@pipeline.command(context_settings=CONTEXT_SETTINGS)
@click.option('--migration', type=click.STRING, help='Run a migration before deploying')
@click.option('--constraint', type=click.STRING, multiple=True, help='Constrain migration and deploy hooks to run on nodes matching labels in the form of "key=value"')
@click.option('--git-remote', type=click.STRING, help='Push deployment tags to git remote')
@click.option('-t', '--timeout', type=click.INT, default=600, help="Timeout deployment rollout after N seconds (default 600)")
@click.option('-u', '--update-config', type=click.BOOL, is_flag=True, help='Also update Kubernetes config')
@click.option('-f', '--filename', type=click.STRING, help='If updating config, use the Kubernetes Yaml file in the ./hokusai directory (default production.yml)')
@click.option('-v', '--verbose', type=click.BOOL, is_flag=True, help='Verbose output')
def promote(migration, constraint, git_remote, timeout, update_config, filename, verbose):
  """Update the project's deployment(s) on production with the image tag
  currently deployed on staging and update the production tag
  to reference the same image"""
  set_verbosity(verbose)
  command(
    hokusai.promote,
    migration,
    constraint,
    git_remote,
    timeout,
    update_config=update_config,
    filename=filename
  )



import os

from jinja2 import Environment, FileSystemLoader, StrictUndefined
from jinja2.exceptions import TemplateError

from hokusai.lib.exceptions import HokusaiError

class TemplateRenderer:
  def __init__(self, template_path, template_config):
    self.template_path = template_path
    self.template_config = template_config

  def load_template(self):
    try:
      env = Environment(loader=FileSystemLoader(os.path.split(self.template_path)[0]), undefined=StrictUndefined)
      return env.get_template(os.path.split(self.template_path)[1])
    except IOError:
      raise HokusaiError("Template %s not found." % self.template_path)

  def render(self):
    try:
      return self.load_template().render(**self.template_config)
    except TemplateError as e:
      raise HokusaiError("Rendering template raised error %s <message '%s'>" % (e.__class__, e.message))



import click

import hokusai

from hokusai.cli.base import base
from hokusai.lib.command import command
from hokusai.lib.common import set_verbosity, CONTEXT_SETTINGS

@base.group()
def registry(context_settings=CONTEXT_SETTINGS):
  """Interact with the project registry defined by `./hokusai/config.yml`"""
  pass


@registry.command(context_settings=CONTEXT_SETTINGS)
@click.option('--tag', type=click.STRING, help='The remote tag to push to (default: the value of `git rev-parse HEAD`)')
@click.option('--local-tag', type=click.STRING, default='latest', help='The local tag to push (default: latest)')
@click.option('--build/--no-build', default=True, help='Force a build of the :latest image before pushing (default: true)')
@click.option('-f', '--filename', type=click.STRING, help='Use the given docker-compose Yaml file when building (default: ./hokusai/build.yml)')
@click.option('--force', type=click.BOOL, is_flag=True, help='Push even if working directory is not clean')
@click.option('--overwrite', type=click.BOOL, is_flag=True, help='Push even if the tag already exists')
@click.option('--skip-latest', type=click.BOOL, is_flag=True, help="Don't update the 'latest' tag" )
@click.option('-v', '--verbose', type=click.BOOL, is_flag=True, help='Verbose output')
def push(tag, local_tag, build, filename, force, overwrite, skip_latest, verbose):
  """Build and push an image to the project's registry tagged as the git SHA1 of HEAD"""
  set_verbosity(verbose)
  command(
    hokusai.push_image,
    tag,
    local_tag,
    build,
    filename,
    force,
    overwrite,
    skip_latest
  )


@registry.command(context_settings=CONTEXT_SETTINGS)
@click.option('--tag', type=click.STRING, default='latest', help='The remote tag to pull  (default: latest)')
@click.option('--local-tag', type=click.STRING, default='latest', help='The local tag to pull to (default: latest)')
@click.option('-v', '--verbose', type=click.BOOL, is_flag=True, help='Verbose output')
def pull(tag, local_tag, verbose):
  """Pull the image tag --tag from the project's registry and tag as --local-tag"""
  set_verbosity(verbose)
  command(
    hokusai.pull,
    tag,
    local_tag
  )


@registry.command(context_settings=CONTEXT_SETTINGS)
@click.option('-e', '--tag-exists', type=click.STRING, help='Exit 0 if the given image tag is found and 1 if not')
@click.option('-r', '--reverse-sort', type=click.BOOL, is_flag=True, help='Sort oldest to latest')
@click.option('-l', '--limit', type=click.INT, default=20, help="Limit output to N images")
@click.option('-f', '--filter-tags', type=click.STRING, help='Filter images that have at least one tag matching the provided string')
@click.option('-d', '--digests', type=click.BOOL, is_flag=True, help='Print image digests')
@click.option('-v', '--verbose', type=click.BOOL, is_flag=True, help='Verbose output')
def images(tag_exists, reverse_sort, limit, filter_tags, digests, verbose):
  """Print images and tags in the project's registry"""
  set_verbosity(verbose)
  command(
    hokusai.images,
    tag_exists,
    reverse_sort,
    limit,
    filter_tags,
    digests
  )

@registry.command(context_settings=CONTEXT_SETTINGS)
@click.argument('tag1', type=click.Choice(['staging', 'production']))
@click.argument('tag2', type=click.STRING)
@click.option('-v', '--verbose', type=click.BOOL, is_flag=True, help='Verbose output')
def retag(tag1, tag2, verbose):
  """
  On the registry, make tag1 point to the same image pointed to by tag2.
  Tag1 must be either 'staging' or 'production'.
  """
  set_verbosity(verbose)
  command(
    hokusai.retag,
    tag1,
    tag2
  )



import os
import sys

import hokusai

# detect whether code is run from a PyInstaller-created bundle
# if is, support files are found in paths as coded in PyInstaller spec file (see 'datas' arg of Analysis() call)
def frozen():
  '''
  return sys.frozen's value if the attribute exists
  else return False
  '''
  return getattr(sys, 'frozen', False)

def cert_file_path():
  ''' return path of cert in PyInstaller bundle '''
  if frozen():
    return sys._MEIPASS + '/lib/cert.pem'

def templates_dir_path():
  '''
  return path of templates in PyInstaller bundle,
  or path in repo, if not PyInstaller bundle
  '''
  if frozen():
    return sys._MEIPASS + '/hokusai_datas/templates'
  else:
    return os.path.dirname(hokusai.__file__) + '/templates'

def version_file_path():
  if frozen():
    return sys._MEIPASS + '/hokusai_datas/VERSION'
  else:
    return os.path.dirname(hokusai.__file__) + '/VERSION'
